from rosemary.model import *
from wooly.forms import *
from wooly.resources import *
from wooly.widgets import *

from formats import *
from main import *
from parameters import *
from util import *

from wooly.widgets import Link

log = logging.getLogger("cumin.task")
strings = StringCatalog(__file__)

class Task(object):
    def __init__(self, app):
        self.app = app
        self.name = self.__class__.__name__
        self.form = None

        self.app.tasks.append(self)

    def init(self):
        log.info("Initializing %s", self)

        assert self.form, self
        self.app.form_page.modes.add_mode(self.form)
        
    def get_title(self, session, obj):
        pass

    def get_description(self, session, obj=None):
        return self.get_title(session, obj)

    def get_href(self, session, obj):
        return self.enter(session, obj).marshal()

    def enter(self, session, obj):
        form_session = wooly.Session(self.app.form_page)

        if obj:
            self.form.id.set(form_session, obj._id)

        self.form.return_url.set(form_session, session.marshal())
        self.form.show(form_session)

        self.do_enter(session, obj, form_session)

        return form_session

    def do_enter(self, session, obj, form_session):
        pass

    def exit(self, session, obj):
        log.debug("Exiting %s", self)

        url = self.form.return_url.get(session)
        osession = wooly.Session.unmarshal(self.app, url)

        self.do_exit(osession, obj)

        log.info("Exited %s", self)

        return osession

    def do_exit(self, session, obj):
        pass

    def exit_with_redirect(self, session, obj):
        osession = self.exit(session, obj)
        self.form.page.redirect.set(session, osession.marshal())

    def start(self, session, obj):
        log.debug("Starting %s", self)

        login = session.client_session.attributes["login_session"]

        invoc = TaskInvocation(self, login)

        now = datetime.now()

        invoc.start_time = now
        invoc.update_time = now
        invoc.status = invoc.PENDING

        log.info("Started %s", self)

        return invoc

    def invoke(self, session, obj, *args, **kwargs):
        if obj:
            assert isinstance(obj, RosemaryObject), obj

        invoc = self.start(session, obj)

        try:
            self.do_invoke(session, obj, invoc, *args, **kwargs)
        except Exception, e:
            self.exception(invoc, e)

    def do_invoke(self, session, obj, invoc, **kwargs):
        pass

    def exception(self, invoc, e):
        now = datetime.now()

        invoc.status = invoc.FAILED
        invoc.end_time = now
        invoc.update_time = now
        invoc.exception = e

        log.exception(e)

    def __str__(self):
        return "%s.%s" % (self.__module__, self.__class__.__name__)

class TaskInvocation(object):
    PENDING = "Pending"
    FAILED = "Failed"
    OK = "OK"

    def __init__(self, task, login_session):
        self.task = task

        login_session = login_session
        login_session.task_invocations.append(self)

        self.user = login_session.user

        self.start_time = None
        self.end_time = None
        self.update_time = None

        self.status = None
        self.exception = None
        self.description = None

        self.status_code = None
        self.output_args = None

        self.dismissed = False

    def get_description(self, session):
        if self.description:
            return self.description

        return self.task.get_description(session)

    def get_status(self, session):
        if self.exception:
            return "%s (%s)" % (self.status, str(self.exception))

        #return "%s: %s" % (self.status, str(self.output_args))
        if self.status is self.OK:
            if self.status_code:
                return self.status_code

        return self.status

    def end(self):
        log.debug("Ending %s", self.task)

        if self.status is self.PENDING:
            self.status = self.OK

        now = datetime.now()

        self.end_time = now
        self.update_time = now

        log.info("Ended %s", self.task)

    @classmethod
    def status_and_args(cls, *args):
        # This is a convenience function that can be
        # used to check results from a call when using
        # custom callbacks instead of completion() below...
        output_args = None
        if len(args) == 0:
            # We have to test this case now since we've made args
            # variable length.  Just call it OK
            status_code = TaskInvocation.OK

        elif len(args) == 1:
            # Status is ok if status_code is not an exception and result
            # goes in output_args
            status_code = args[0]
            if not isinstance(status_code, Exception):
                output_args = status_code
                status_code = TaskInvocation.OK
        else:
            status_code, output_args = args[0:2]
        return status_code, output_args

    @classmethod
    def is_success(cls, *args):
        s, a = TaskInvocation.status_and_args(*args)
        return s in (0, TaskInvocation.OK)

    def make_callback(self):
        def completion(*args):
            # Callback argument formats come in two basic flavors in Cumin
            # depending on the semantics of the async op:
            # callback(status, result)
            # callback(result) where type(result) == Exception indicates
            #   a failure status.
            # Allow this general mechanism to handle both types.
            status_code, output_args = self.status_and_args(*args)

            # Make results recorded directly through the callback
            # match what is done when TaskInvocation values are
            # set by an ObjectTask or Task on failure
            # (see Task.invoke() and Task.exception() above)
            # Previously the different mechanisms produced different results.
            if isinstance(status_code, Exception):
                self.status = self.FAILED
                self.exception = status_code
            else:
                self.status_code = status_code
                self.output_args = output_args
            self.end()
        return completion

class TaskInvocationSet(Widget):
    def __init__(self, app, name):
        super(TaskInvocationSet, self).__init__(app, name)

        self.update_enabled = True

        self.dismiss = IntegerParameter(app, "dismiss")
        self.add_parameter(self.dismiss)
        
        self.dismiss_all = BooleanParameter(app, "dismiss_all")
        self.add_parameter(self.dismiss_all)

        self.item_widget = TaskInvocationItem(app, "item")
        self.add_child(self.item_widget)

    def get_items(self, session):
        items = []
        if("login_session" in session.client_session.attributes.keys() ):
            login = session.client_session.attributes["login_session"]
            items = [x for x in login.task_invocations if not x.dismissed]
        return items

    def do_process(self, session):
        super(TaskInvocationSet, self).do_process(session)

        dismiss = self.dismiss.get(session)
        dismiss_all = self.dismiss_all.get(session)
        
        if(dismiss_all):
            for invoc in self.get_items(session):
                invoc.dismissed = True
            self.dismiss_all.set(session, self.dismiss_all.default)
                    
        if dismiss:
            for invoc in self.get_items(session):
                if id(invoc) == dismiss:
                    invoc.dismissed = True
                    break
            self.dismiss.set(session, self.dismiss.default)


    def do_render(self, session):
        items = self.get_items(session)

        if items:
            return super(TaskInvocationSet, self).do_render(session)

    def render_items(self, session):
        writer = Writer()

        for item in self.get_items(session):
            writer.write(self.item_widget.render(session, item))

        return writer.to_string()
    
    def render_dismiss_all_href(self, session):
        branch = session.branch()
        self.dismiss_all.set(branch, True)
        return branch.marshal()

class TaskInvocationItem(Widget):
    def render_icon_href(self, session, item):
        return "resource?name=add-20.png"

    def render_message(self, session, item):
        description = item.get_description(session)
        description = xml_escape(description)

        if not description:
            description = ""

        status = xml_escape(str(item.get_status(session)))

        return "%s: %s" % (description, status)

    def render_dismiss_href(self, session, item):
        branch = session.branch()
        self.parent.dismiss.set(branch, id(item))
        return branch.marshal()

class TaskLink(Link):
    def __init__(self, app, name, task, object=None):
        super(TaskLink, self).__init__(app, name)

        self.task = task
        self.object = object

    def render_href(self, session):
        obj = None

        if self.object:
            obj = self.object.get(session)

        return self.task.get_href(session, obj)

    def render_content(self, session):
        obj = None

        if self.object:
            obj = self.object.get(session)

        return self.task.get_title(session, obj)

class TaskButton(FormButton):
    def __init__(self, app, name, task, object=None):
        super(TaskButton, self).__init__(app, name)

        self.task = task
        self.object = object

    def process_submit(self, session):
        obj = None

        if self.object:
            obj = self.object.get(session)

        href = self.task.get_href(session, obj)
        self.page.redirect.set(session, href)

    def render_content(self, session):
        obj = None

        if self.object:
            obj = self.object.get(session)

        return self.task.get_title(session, obj)

class ObjectTaskForm(FoldingFieldSubmitForm):
    def __init__(self, app, name, task, cls):
        super(ObjectTaskForm, self).__init__(app, name)

        self.task = task
        self.cls = cls

        self.id = IntegerParameter(app, "id")
        self.add_parameter(self.id)

        self.object = ObjectAttribute(app, "object", self.cls, self.id)
        self.add_attribute(self.object)

    def do_process(self, session):
        self.object.process(session)

        super(ObjectTaskForm, self).do_process(session)

    def process_submit(self, session):
        self.validate(session)

        if not self.errors.get(session):
            obj = self.object.get(session)

            self.task.invoke(session, obj)
            self.task.exit_with_redirect(session)

    def render_title(self, session):
        obj = self.object.get(session)
        return self.task.get_title(session, obj)

    def render_content(self, session):
        if len(self.main_fields.fields):
            return super(ObjectTaskForm, self).render_content(session)

        obj = self.object.get(session)
        return obj.get_title()

# class SelectionTaskForm
